#!/usr/bin/python
# -*-coding:Utf-8-*
#
# 
# Additional temperature sensors for the RECS platform at IRIT.
# We make use of up to 16 x TCN75A sensors from Microchip.
# Sensors are dispatched within both i2C buses
#
# François Septembre 2013
#
# Nov.13: added --calibrate option to compute the per sensor compensation factor
#


# #############################################################################
#
# Import zone
#
import errno
import os
import signal
import syslog
import sys

#import math
from math import fsum
from optparse import OptionParser
import string
import struct
import time
from datetime import datetime
#import datetime
#from threading import Thread, Lock, Event
import threading

import socket
#from socket import socket, AF_INET, SOCK_STREAM, SHUT_RDWR
import json

# I2C management & devices
from Adafruit_I2C import Adafruit_I2C
#import I2C_TCN75A
import enable_second_I2C
#from enable_second_I2C import i2cConfig



# #############################################################################
#
# Global Variables
#

# Description of the various sensors along with their base address
# Format is of each element is [ i2cBus, i2cDeviceAddr, recsID, correction_factor ]
# exemple: [0, 0x48, 9, 0.985] = recs9, i2cbus0 (P5) addr=0x48 and value=I2C_value*0.985
# Note: we use both I2C buses, the second one, I2CBUS=0, requires some I/O remapping
SENSORS_LIST = [[0,0x48,9],[0,0x4A,18],[0,0x4B,10],[1,0x48,1],[1,0x49,2],[1,0x4A,3],[1,0x4B,4],[1,0x4C,5],[1,0x4D,6],[1,0x4E,7],[1,0x4F,8]]
CFACTORS = [1.0267124572178723, 1.0090194903930332, 0.9777973745311662, 0.9901991815694875, 0.9757605462788651, 0.9693465390361313, 0.9862746864233116, 1.0223625392453226, 1.0185701993069607, 1.0256545360873026, 1.0029196244561482]

# Zabbix related
HOST_PREFIX="recs"
HOST_SUFFIX=".coolemall.irit.fr"

zabbix_server = "frontal.coolemall.irit.fr"
zabbix_serverport = 9999
zabbix_key = "ext_temperature"

# Measurement related
measure_interleave = 1.0
measures_flush_nb = 45

I2Cmanagers = []	# liste des instances de la classe I2Cmanager
I2CmanagersDelay = 3.5	# délai de démarrage des acquisitions entre les deux bus I2C,
                        # pour étaler les envois au serveur Zabbix
SensorsCalibration = False
__shutdown = False


'''
TO IMPLEMENT HERE !!!
# Almost real-time threading
import threading;

def do_every (interval, worker_func, iterations = 0):
  if iterations != 1:
    threading.Timer (
      interval,
      do_every, [interval, worker_func, 0 if iterations == 0 else iterations-1]
    ).start ();

  worker_func ();

def print_hw ():
  print "hello world";

def print_so ():
  print "stackoverflow"


# call print_so every second, 5 times total
do_every (1, print_so, 5);

# call print_hw two times per second, forever
do_every (0.5, print_hw);
'''


#
# Class TCN75A
#	Temperature sensor from Microchip
#
class tcn75a:

	#
	# TCN75A registers
	REG_TA			= 0x00	# 2 bytes signed
	REG_CONFIG		= 0x01
	REG_THYST		= 0x02	# 2 bytes signed
	REG_TSET		= 0x03	# 2 bytes signed

	def __init__(self, i2cbus, i2cdeviceaddr, cfactor=1.0, debug=False):
		self.i2cbus = i2cbus
		self.debug = debug
		self.i2cdevaddr = i2cdeviceaddr
		self.cfactor = cfactor
		self.measures = []
		self.data2send = []
		self._sensor = -1
		self.sensormeanvalue = 0.0

		# I2C device initalisation (last parameter is DEBUG ==> True/False)
		self._sensor = Adafruit_I2C(i2cdeviceaddr, i2cbus, debug);
	
		# Configure sensor for 11 bits resolution (0.125°c) 0x40
		# Configure sensor for 10 bits resolution (0.25°c) 0x20
		cur_config = 0x40;
		res = self._sensor.write8(self.REG_CONFIG,cur_config);
		if res == -1:
			# device does not exists :(
			raise IOError

	def acquire(self, date):
		# let's read the current temperature
		cur_temp = self._sensor.readList(self.REG_TA,2)
		# Value computation for 10 bits resolution (0.25°c) 0x20
		#_value = float(cur_temp[0]) + float((cur_temp[1]>>6)*25)/100
		# Value computation for 11 bits resolution (0.125°c) 0x40
		_value = float(cur_temp[0]) + float((cur_temp[1]>>5)*125)/1000
#		print "Current temperature is %02d.%03d°c ... float=%.03f°c\n" % (cur_temp[0], (cur_temp[1]>>5)*125, _value);
		# Apply correction factor if not in calibration process
		_value = round( _value*(float(self.cfactor) if SensorsCalibration==False else 1.0), 3)
#		print "Corrected value = %.03f°c\n" % (_value)
		# store value with timestamp
		self.measures.append([_value, date])
		return _value

	def savedata(self):
		# when it's time to send values to zabbix,
		# we copy current values to another attribute
		# to avoid race conditions with newer values arriving
		# Note: this is a *real* copy, not just a new reference
		self.data2send = self.measures[:]
		self.measures = []
		# Mean Value computation intended to calibration purpose
		if SensorsCalibration==True:
			self.sensormeanvalue = fsum([val for val,time in self.data2send]) / len(self.data2send)
			print "Device 0x%02x mean value = %.03f°c" % (self.i2cdevaddr,self.sensormeanvalue) 
#		print "Device 0x%02x saved %d values" % (self.i2cdevaddr,len(self.data2send))
		return self.sensormeanvalue

	def genJSON(self,zabbix_host):
		"Generate JSON body with measured values"
		jsonBody = ''
		i = 0
		for i in range(len(self.data2send)):
			jsonBody += ' { "host":"'+zabbix_host+'", "key":"'+zabbix_key+'", "value":"'+str(self.data2send[i][0])+'", "clock":'+str(self.data2send[i][1])+' }'

			if (i < len(self.data2send)-1):
				jsonBody += ',\n'

		self.data2send = []
		# Send back JSON body + number of data
		return jsonBody, i+1

# End class TCN75A



#
# Class I2Cmanager
#	used to manage configuration and data acquisition of various
#	sensors connected on an I2C bus.
#
class I2Cmanager:
	def __init__(self, i2cbus=1, debug=False):
		self.i2cbus = i2cbus
		self.debug = debug
		self._sensors = []
		self._nb_acquire = 0
		self._timer = -1

		if ( self.i2cbus==0):
			# activate second i2Cbus
			print "\t... activating I2C0, the second I2C bus on RPi rev. 2.0  ..."
			enable_second_I2C.i2cConfig()

	# Function to add an i2c device
	def addSensor(self, i2caddr, recsid, cfactor):
		# instantiate the sensor
		try:
			_sensor = tcn75a( self.i2cbus, i2caddr, cfactor, self.debug )
			# add sensor to list with its recsID
			self._sensors.append([_sensor,recsid])
		except IOError:
			# device probably does not exist
			pass

	# Function to collect average value of all sensors
	def getavg(self, avgValues):
		for sensor in range(len(self._sensors)):
			avgValues.append(self._sensors[sensor][0].sensormeanvalue)

    # Function to start whole i2c data acquisition ...
	def start(self):
#		global measure_interleave
		# set timer to call function _measure after interval specified
		self._timer = threading.Timer(measure_interleave, self._measure)
		# ... and launch it
		self._timer.start()

	# Function handeling metric collection
	def _measure(self):

		# ... and immediately relaunch the timer
		self._timer = threading.Timer(measure_interleave, self._measure)
		self._timer.start()

		# empty list ?
		if len(self._sensors) == 0:
			print "I2Cmanager '%d' has no devices!" % (self.i2cbus)
			return

		# check time it takes to collect all measures
#		t0 = datetime.now()

		# to timestamp the datas
		date = time.time()

		# collect data from sensors in our list
		for sensor in range(len(self._sensors)):
			# select current sensor
			cur_sensor,recsID = self._sensors[sensor]
			# try to read the value
			try:
				_value = cur_sensor.acquire(date)
				print "RecsID %d: %0.3f°c" % (recsID,_value)

			except IOError:
				# exception due to an I2C bus transaction error
				print "###ERROR reading a device!"

#		t1=datetime.now()
#		elapsed_time=t1 - t0
#		print "[I2C%d] acquisitions took %dms\n" % (self.i2cbus, elapsed_time.microseconds/1000);

		# increment nb acquisitions
		self._nb_acquire += 1

		# send data if sufficient number acquired
		if (self._nb_acquire >= measures_flush_nb) :
			print "[I2C%d] Reached %d measures ... send them to zabbix :)" % (self.i2cbus,measures_flush_nb)

			# first reset iterations number
			self._nb_acquire = 0
			# ... then all sensors ought to save their current values (to avoid races conditions)
			self._saveData()
			# ... and now send all sensors' data to Zabbix
			self._sendData()

	# Function to order sensors to save data that will be sent to zabbix
	def _saveData(self):
		_sensormeanvalue = 0.0
		# All sensors copy measures in internal class attribute
		for sensor in range(len(self._sensors)):
			cur_sensor = self._sensors[sensor][0]
			_sensormeanvalue=cur_sensor.savedata()

	# Function to send all acquired datas to Zabbix
	def _sendData(self):
		"Send all sensors' data to Zabbix"
		_total_values = 0
		# generate JSON frame
		jsonHeader = '{ "request":"sender data",\n"data":[\n'
		jsonBody = '';
		jsonFooter = '\n],\n"clock":'+str(time.time())+' }'

		# Creating the json request body with measured values
		for sensor in range(len(self._sensors)):
			# 0 --> TCN75A sensor
			# 1 --> recsID
			cur_sensor = self._sensors[sensor][0]
			_jsonBody,_nb = cur_sensor.genJSON(HOST_PREFIX+str(self._sensors[sensor][1])+HOST_SUFFIX)
			jsonBody += _jsonBody
			_total_values += _nb

			if (sensor < len(self._sensors)-1):
				jsonBody += ',\n'

		# DEBUG
#		print jsonBody
		print "[I2C%d] send %d values to Zabbix ..." % (self.i2cbus,_total_values)
		# DEBUG

		# Send JSON frame to Zabbix
		if ( _total_values > 0 ):
			zabbix_sendData(jsonHeader+jsonBody+jsonFooter)
		else:
			print "[I2C%d] No data to send to zabbix!" % (self.i2cbus)

	# Function to stop everything
	def stop(self):
		# stop measures
		self._timer.cancel()
		print "[I2C%d] Stopping external temperature sensors acquisition" % (self.i2cbus)
		# save all data from each sensor
		self._saveData()
		# ... and send them to zabbix
		self._sendData()

# End class I2Cmanager


# Function sending data to the zabbix server
def zabbix_sendData(jsonFrame):
	"Sending JSON frame to Zabbix server"

	zabbix_header = '''ZBXD\1%s%s'''
	zabbix_data = jsonFrame

	data_length = len(zabbix_data)
	data_header = struct.pack('i', data_length) + '\0\0\0\0'

	data_to_send = zabbix_header % (data_header, zabbix_data)

	# Initializing socket
	try:
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		sock.settimeout(3)
		sock.connect((zabbix_server, zabbix_serverport))
		sock.settimeout(None)
	except Exception as e:
		print "zabbix_sendData, socket initialization error : ", e
#		raise
#		stopMonitoring()
		return -1

	# Sending data to the server
	try:
		sock.send(data_to_send)
	except Exception as e:
		print "zabbix_sendData, socket sending data error : ", e
#		raise
#		stopMonitoring()
		return -1

	# Receiving server response
	try:
		response_header = sock.recv(5)
		if not response_header == 'ZBXD\1':
			print "Zabbix response is not valid ?!?!"
			raise ValueError('Got invalid response')

		# read the data header to get the length of the response
		response_data_header = sock.recv(8)
		response_data_header = response_data_header[:4] # we are only interested in the first four bytes
		response_len = struct.unpack('i', response_data_header)[0]

		# read the whole rest of the response now that we know the length
		response_raw = sock.recv(response_len)
	except Exception as e:
		print "zabbix_sendData, socket receiving data error : ", e
		raise
		stopMonitoring()

	sock.close()

	#response = json.loads(response_raw)
	#print response



#
# Function ctrlc_handler
def ctrlc_handler(signum, frame):
	global __shutdown
	__shutdown = True
	stopMonitoring()

#
# Function stoping the monitoring
def stopMonitoring():
	global I2Cmanagers
	for i in range(len(I2Cmanagers)):
		I2Cmanagers[i].stop()
	del I2Cmanagers


# #############################################################################
#
# MAIN
#

def main():
	global I2Cmanagers,I2CmanagersDelay,SensorsCalibration,measures_flush_nb

	# Trap CTRL+C (kill -2)
	signal.signal(signal.SIGINT, ctrlc_handler)

	# Parse arguments
	parser = OptionParser()
	parser.add_option("-c", "--calibrate",
						action="store_true", dest="calibrate", default=False,
						help="Start calibration process")
	(options, args) = parser.parse_args()
	
	# Arguments evaluation
	if options.calibrate==True:
		SensorsCalibration=options.calibrate
		I2CmanagersDelay=(3.5 if SensorsCalibration==False else 0.0)
		#measures_flush_nb=5
		print "\n### CALIBRATION of sensors activated ... disabling current correction factors"
		time.sleep(1)

	# Register I2CBuses
	# note: for I2CBus=0 (Raspberry Pi model B latest revision), we ought to modify I/O mapping
	for i in range(2):
		#_i2cmanager = I2Cmanager(i, True)
		_i2cmanager = I2Cmanager(i, False)
		I2Cmanagers.append(_i2cmanager)

	# Register specified sensors
	for i in range(len(SENSORS_LIST)):
		# retrieve sensor IDs
		i2cbus,i2caddr,recsID = SENSORS_LIST[i]
		cfactor = CFACTORS[i]
		print "i=%d, i2cbus=%d, i2caddr=0x%02X, recsID=%d, cfactor=%.03f" % (i,i2cbus,i2caddr, recsID, cfactor if SensorsCalibration==False else 1.0)
		# add sensor to corresponding I2Cmanager		
		(I2Cmanagers[i2cbus]).addSensor(i2caddr,recsID,cfactor)

	# Launch data acquisition
	for i,manager in enumerate(I2Cmanagers):
		manager.start()
		# sleep to avoid all I2Cmanagers to send whole data to Zabbix at the same time
		time.sleep(I2CmanagersDelay)

	# Main loop
	time.sleep(1)
	while SensorsCalibration==True and __shutdown==False:
		time.sleep(measures_flush_nb)
		# compute the global average value		
		avgGlobalValues = []
		for i,manager in enumerate(I2Cmanagers):
			manager.getavg(avgGlobalValues)
		avg=fsum(avgGlobalValues)/len(avgGlobalValues)
		print "\n### Calibration ..."
		print avgGlobalValues
		print "Min < Avg < Max --> %.03f < %.03f < %.03f" % (min(avgGlobalValues), avg, max(avgGlobalValues))
		cfactors=[avg/val for val in avgGlobalValues]
		print "\n>>> correction factors to be applied :"
		print cfactors
		print "\n"

# Execution or import
if __name__ == "__main__":

	# Start executing
	main()


# The END - Jim Morrison 1943 - 1971
#sys.exit(0)





###
### FIN
###

